# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: health.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0chealth.proto\x12\x11\x63oncordium.health\"\x13\n\x11NodeHealthRequest\"\x14\n\x12NodeHealthResponse2^\n\x06Health\x12T\n\x05\x43heck\x12$.concordium.health.NodeHealthRequest\x1a%.concordium.health.NodeHealthResponseB/\n\x16\x63om.concordium.grpc.v2P\x01\xaa\x02\x12\x43oncordium.Grpc.V2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'health_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.concordium.grpc.v2P\001\252\002\022Concordium.Grpc.V2'
  _globals['_NODEHEALTHREQUEST']._serialized_start=35
  _globals['_NODEHEALTHREQUEST']._serialized_end=54
  _globals['_NODEHEALTHRESPONSE']._serialized_start=56
  _globals['_NODEHEALTHRESPONSE']._serialized_end=76
  _globals['_HEALTH']._serialized_start=78
  _globals['_HEALTH']._serialized_end=172
# @@protoc_insertion_point(module_scope)
